import unittest
from .utils import PaspTest
import numpy as np
import pasp

class TestExamples(PaspTest):
  def test_asia(self):
    P = pasp.parse("examples/asia.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(trip)
    self.assertApproxEqual(R[0,:], [0.01, 0.01])
    # ℙ(tuberculosis | trip)
    self.assertApproxEqual(R[1,:], [0.05, 0.05])
    # ℙ(cancer | smoking)
    self.assertApproxEqual(R[2,:], [0.1,  0.1 ])
    # ℙ(test | or)
    self.assertApproxEqual(R[3,:], [0.98, 0.98])
    # ℙ(smoking)
    self.assertApproxEqual(R[4,:], [0.5,  0.5 ])
    # ℙ(tuberculosis | not trip)
    self.assertApproxEqual(R[5,:], [0.01, 0.01])
    # ℙ(cancer | not smoking)
    self.assertApproxEqual(R[6,:], [0.01, 0.01])
    # ℙ(test | not or)
    self.assertApproxEqual(R[7,:], [0.05, 0.05])

  def test_earthquake(self):
    P = pasp.parse("examples/earthquake.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(alarm | burglary, earthquake)
    self.assertApproxEqual(R[0,:], [0.9,  0.9])
    # ℙ(alarm | not burglary, earthquake)
    self.assertApproxEqual(R[1,:], [0.1,  0.1])
    # ℙ(alarm | burglary, not earthquake)
    self.assertApproxEqual(R[2,:], [0.8,  0.8])
    # ℙ(alarm | not burglary, not earthquake)
    self.assertApproxEqual(R[3,:], [0.0,  0.0])

  def test_game(self):
    P = pasp.parse("examples/game.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(wins(b))
    self.assertApproxEqual(R[0,:], [0.7,  1.0])
    # ℙ(wins(c))
    self.assertApproxEqual(R[1,:], [0.3,  0.3])

  def test_insomnia(self):
    P = pasp.parse("examples/insomnia.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(insomnia)
    self.assertApproxEqual(R[0,:], [0.3,  0.3])
    # ℙ(work)
    self.assertApproxEqual(R[1,:], [0.3,  1.0])
    # ℙ(sleep)
    self.assertApproxEqual(R[2,:], [0.0,  0.7])
    # ℙ(not sleep)
    self.assertApproxEqual(R[3,:], [0.3,  1.0])
    # ℙ(not work)
    self.assertApproxEqual(R[4,:], [0.0,  0.7])

  def test_prisoners(self):
    P = pasp.parse("examples/prisoners.lp")
    R = pasp.exact(P, quiet = True)
    α = 19/40
    # ℙ(e1 | u)
    print(R[0,:], [1.0/(1+2*(((1-α)/α)**2)), 1.0/(1+2*((α/(1-α))**2))])
    self.assertApproxEqual(R[0,:], [1.0/(1+2*(((1-α)/α)**2)), 1.0/(1+2*((α/(1-α))**2))])
    # ℙ(e1 | not b, u)
    self.assertApproxEqual(R[1,:], [1.0/(1+((1-α)/α)**2), 1.0/(1+((α/(1-α))**2))])
    # ℙ(g | e1, u)
    self.assertApproxEqual(R[2,:], [0.0, 1.0])
    # ℙ(d)
    self.assertApproxEqual(R[3,:], [0.0, 1.0])
    # ℙ(e1 | g, u)
    self.assertApproxEqual(R[4,:], [0.0, 1.0/(1+(α/(1-α))**2)])
    # ℙ(e1 | ga, u)
    self.assertApproxEqual(R[5,:], [1.0/(1+(((1-α)/α)**2)/α), 1.0/(1+((α/(1-α))**2)/(1-α))])

  def test_simple(self):
    P = pasp.parse("examples/simple.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(s(a))
    self.assertApproxEqual(R[0,:], [0.20, 0.20])
    # ℙ(s(b))
    self.assertApproxEqual(R[1,:], [0.30, 0.30])
    # ℙ(v)
    self.assertApproxEqual(R[2,:], [0.048, 0.048])

  def test_simpler(self):
    P = pasp.parse("examples/simpler.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(r)
    self.assertApproxEqual(R[0,:], [0.5, 0.5])
    # ℙ(v)
    self.assertApproxEqual(R[1,:], [0.25, 0.25])

  def test_smokers(self):
    P = pasp.parse("examples/smokers.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(smokes(a))
    self.assertApproxEqual(R[0,:], [0.06, 0.06])
    # ℙ(smokes(b))
    self.assertApproxEqual(R[1,:], [0.2, 0.2])

  def test_earthquake_ad(self):
    P = pasp.parse("examples/earthquake_ad.lp")
    R = pasp.exact(P, quiet = True)
    # ℙ(alarm | burglary, earthquake(heavy))
    self.assertApproxEqual(R[0,:], [0.9, 0.9])
    # ℙ(alarm | not burglary, earthquake(mild))
    self.assertApproxEqual(R[1,:], [0.1, 0.1])
    # ℙ(alarm | burglary, not earthquake(mild))
    self.assertApproxEqual(R[2,:], [0.8058823529411767, 0.8058823529411767])
    # ℙ(alarm | not burglary, earthquake(none))
    self.assertApproxEqual(R[3,:], [0.0, 0.0])

  def test_multinsomnia(self):
    P = pasp.parse("examples/multinsomnia.lp")
    R = pasp.exact(P, quiet = True)
    O = [[0.300000, 0.300000], [0.500000, 0.500000], [0.700000, 0.700000], [0.300000, 1.000000],
         [0.500000, 1.000000], [0.700000, 1.000000], [0.000000, 0.700000], [0.000000, 0.500000],
         [0.000000, 0.300000], [0.030000, 0.200000], [0.042000, 0.200000], [0.067500, 0.450000],
         [0.157500, 0.450000], [0.073500, 0.350000], [0.122500, 0.350000]]
    for i, o in enumerate(O):
      self.assertApproxEqual(R[i,:], o)

  def test_fault_tree(self):
    P = pasp.parse("examples/fault_tree.lp")
    R = pasp.exact(P, quiet = True)
    # All marginal probabilities.
    T = np.array([[0.05     , 0.1       ], [0.05      , 0.1       ], [0.005    , 0.01      ],
                  [0.05475  , 0.109     ], [0.05475   , 0.109     ], [0.01198  , 0.019925  ],
                  [0.0074875, 0.0199    ], [0.01741262, 0.0346015 ], [0.0193778, 0.03942849],
                  [0.001    , 0.01      ], [0.02035842, 0.04903421]])
    self.assertTrue(np.allclose(R[0:-4], T))

  # This test might take a long time. Ignore for now.
  # def test_rain_util(self):
    # P = pasp.parse("examples/rain_utility.lp")
    # R = pasp.exact(P, quiet = True)
    # self.assertAlmostEqual(R[0,1], 0.5075)

class TestLStable(PaspTest):
  def test_barber(self):
    P = pasp.parse("examples/barber.lp", semantics = "lstable")
    R = pasp.exact(P, quiet = True)
    # ℙ(shaves(b, a))
    self.assertApproxEqual(R[0,:], [1.0, 1.0])
    # ℙ(not shaves(b, b))
    self.assertApproxEqual(R[1,:], [0.5, 0.5])
    # ℙ(undef shaves(b, b))
    self.assertApproxEqual(R[2,:], [0.5, 0.5])

  def test_3coloring(self):
    P = pasp.parse("examples/3coloring.lp", semantics = "lstable")
    R = pasp.exact(P, quiet = True)
    # ℙ(c(1, r))
    self.assertApproxEqual(R[0,:], [0.0, 1.0])
    # ℙ(e(1, 2) | undef f)
    self.assertApproxEqual(R[1,:], [0.7727272727272727, 0.7727272727272727])
    # ℙ(undef f)
    self.assertApproxEqual(R[2,:], [0.064453125, 0.064453125])

class TestPlog(PaspTest):
  def test_asia(self):
    P = pasp.parse("examples/asia.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(trip)
    self.assertAlmostEqual(R[0], 0.01)
    # ℙ(tuberculosis | trip)
    self.assertAlmostEqual(R[1], 0.05)
    # ℙ(cancer | smoking)
    self.assertAlmostEqual(R[2], 0.1)
    # ℙ(test | or)
    self.assertAlmostEqual(R[3], 0.98)
    # ℙ(smoking)
    self.assertAlmostEqual(R[4], 0.5)
    # ℙ(tuberculosis | not trip)
    self.assertAlmostEqual(R[5], 0.01)
    # ℙ(cancer | not smoking)
    self.assertAlmostEqual(R[6], 0.01)
    # ℙ(test | not or)
    self.assertAlmostEqual(R[7], 0.05)

  def test_earthquake(self):
    P = pasp.parse("examples/earthquake.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(alarm | burglary, earthquake)
    self.assertAlmostEqual(R[0], 0.9)
    # ℙ(alarm | not burglary, earthquake)
    self.assertAlmostEqual(R[1], 0.1)
    # ℙ(alarm | burglary, not earthquake)
    self.assertAlmostEqual(R[2], 0.8)
    # ℙ(alarm | not burglary, not earthquake)
    self.assertAlmostEqual(R[3], 0.0)

  def test_simple(self):
    P = pasp.parse("examples/simple.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(s(a))
    self.assertAlmostEqual(R[0], 0.20)
    # ℙ(s(b))
    self.assertAlmostEqual(R[1], 0.30)
    # ℙ(v)
    self.assertAlmostEqual(R[2], 0.048)

  def test_simpler(self):
    P = pasp.parse("examples/simpler.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(r)
    self.assertAlmostEqual(R[0], 0.5)
    # ℙ(v)
    self.assertAlmostEqual(R[1], 0.25)

  def test_smokers(self):
    P = pasp.parse("examples/smokers.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(smokes(a))
    self.assertAlmostEqual(R[0], 0.06)
    # ℙ(smokes(b))
    self.assertAlmostEqual(R[1], 0.2)

  def test_game(self):
    P = pasp.parse("examples/game.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(wins(b))
    self.assertAlmostEqual(R[0], 1.7/2)
    # ℙ(wins(c))
    self.assertAlmostEqual(R[1], 0.3)

  def test_insomnia(self):
    P = pasp.parse("examples/insomnia.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(insomnia)
    self.assertAlmostEqual(R[0], 0.3)
    # ℙ(work)
    self.assertAlmostEqual(R[1], 0.3+(1.0-0.3)/2)
    # ℙ(sleep)
    self.assertAlmostEqual(R[2], 0.7/2)
    # ℙ(not sleep)
    self.assertAlmostEqual(R[3], 0.3+(1.0-0.3)/2)
    # ℙ(not work)
    self.assertAlmostEqual(R[4], 0.7/2)

  def test_earthquake_ad(self):
    P = pasp.parse("examples/earthquake_ad.lp")
    R = pasp.exact(P, psemantics = "maxent", quiet = True).flatten()
    # ℙ(alarm | burglary, earthquake(heavy))
    self.assertAlmostEqual(R[0], 0.9)
    # ℙ(alarm | not burglary, earthquake(mild))
    self.assertAlmostEqual(R[1], 0.1)
    # ℙ(alarm | burglary, not earthquake(mild))
    self.assertAlmostEqual(R[2], 0.8058823529411767)
    # ℙ(alarm | not burglary, earthquake(none))
    self.assertAlmostEqual(R[3], 0.0)

if __name__ == "__main__":
  unittest.main()



import unittest

from .utils import PaspTest, hoeffding
import pasp
import numpy as np

N_SAMPLES  = 500
N_ITERS    = 100
"""Epsilon for all test cases based on the number of samples. Give some tolerance, since it's an
approximation of an approximation."""
EPS = hoeffding(N_SAMPLES)*1.5

class TestLearning(PaspTest):
  def test_insomnia_ad(self):
    which = "examples/insomnia_ad.lp"
    W = [[0.3, 0.2, 0.5], [0.1, 0.6, 0.3], [0.9, 0.1, 0.0], [0.0, 0.5, 0.5]]
    P = pasp.parse(which)
    Q = pasp.parse(which)
    A = ["work(anna)", "work(bill)", "sleep(anna)", "sleep(bill)"]

    for w in W:
      P.AD[0].P = w
      S = pasp.sample(P, A, n = N_SAMPLES)
      Q.AD[0].P = [1/len(w) for _ in range(len(w))]
      pasp.learn(Q, S, A, niters = N_ITERS)

      for u, v in zip(P.AD[0].P, Q.AD[0].P):
        self.assertAlmostEqual(u, v, delta = EPS)

  def test_earthquake_ad(self):
    which = "examples/earthquake_ad.lp"
    P = pasp.parse(which)
    Q = pasp.parse(which)
    A = ["alarm", "calls(a)", "calls(b)"]
    S = pasp.sample(Q, A, n = N_SAMPLES)
    P.PF[0].p = 0.5; P.PF[0].learnable = True
    pasp.learn(P, S, A, niters = N_ITERS)

    self.assertAlmostEqual(P.PF[0].p, Q.PF[0].p, delta = EPS)

if __name__ == "__main__":
  unittest.main()



import unittest

from .utils import PaspTest, hoeffding
import pasp
import numpy as np
import math

N_SAMPLES  = 2000
"Epsilon for all test cases based on the number of samples."
EPS = hoeffding(N_SAMPLES)

#print("Hoeffding tolerance:", EPS)

class SamplingCases(PaspTest):
  def test_insomnia_ad(self):
    W = [[0.3, 0.2, 0.5], [0.1, 0.6, 0.3], [0.9, 0.1, 0.0], [0.0, 0.5, 0.5]]
    P = pasp.parse("examples/insomnia_ad.lp")
    A = ["insomnia(anna)", "insomnia(bill)", "work(anna)", "work(bill)", "sleep(anna)",
         "sleep(bill)"]
    for w in W:
      P.AD[0].P = w
      R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
      S = pasp.sample(P, A, n = N_SAMPLES)
      Q = np.sum(S, axis = 0) / N_SAMPLES

      self.assertTrue(np.allclose(R, Q, atol = EPS))

  def test_earthquake(self):
    P = pasp.parse("examples/earthquake.lp")
    A = ["alarm", "burglary", "earthquake"]
    R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
    S = pasp.sample(P, A, n = N_SAMPLES)

    # ℙ(alarm | burglary, earthquake)
    p = np.sum(S[:,0] & S[:,1] & S[:,2]) # ℙ(alarm, burglary, earthquake)
    q = np.sum(S[:,1] & S[:,2])          # ℙ(burglary, earthquake)
    self.assertAlmostEqual(p/q, R[0], delta = EPS)

    # ℙ(alarm | ¬burglary, earthquake)
    p = np.sum(S[:,0] & ~S[:,1] & S[:,2]) # ℙ(alarm, ¬burglary, earthquake)
    q = np.sum(~S[:,1] & S[:,2])          # ℙ(¬burglary, earthquake)
    self.assertAlmostEqual(p/q, R[1], delta = EPS)

    # ℙ(alarm | burglary, ¬earthquake)
    p = np.sum(S[:,0] & S[:,1] & ~S[:,2]) # ℙ(alarm, burglary, ¬earthquake)
    q = np.sum(S[:,1] & ~S[:,2])          # ℙ(burglary, ¬earthquake)
    self.assertAlmostEqual(p/q, R[2], delta = EPS)

    # ℙ(alarm | ¬burglary, ¬earthquake)
    p = np.sum(S[:,0] & ~S[:,1] & ~S[:,2]) # ℙ(alarm, ¬burglary, ¬earthquake)
    q = np.sum(~S[:,1] & ~S[:,2])          # ℙ(¬burglary, ¬earthquake)
    self.assertAlmostEqual(p/q, R[3], delta = EPS)

  def test_asia(self):
    P = pasp.parse("examples/asia.lp")
    A = ["trip", "smoking", "tuberculosis", "cancer", "or", "test"]
    R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
    S = pasp.sample(P, A, n = N_SAMPLES)

    # These probabilities are too small for us to reliably approximate conditionals by sampling.
    # Since Hoeffding only holds for the distribution we are sampling from (the joint), we only
    # test against these.

    # ℙ(trip)
    self.assertAlmostEqual(np.sum(S[:,0])/N_SAMPLES, R[0], delta = EPS)

    # ℙ(smoking)
    self.assertAlmostEqual(np.sum(S[:,1])/N_SAMPLES, R[4], delta = EPS)

  def test_game(self):
    P = pasp.parse("examples/game.lp")
    A = ["wins(b)", "wins(c)"]
    R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
    S = pasp.sample(P, A, n = N_SAMPLES)

    # ℙ(wins(b))
    self.assertAlmostEqual(np.sum(S[:,0])/N_SAMPLES, R[0], delta = EPS)

    # ℙ(wins(c))
    self.assertAlmostEqual(np.sum(S[:,1])/N_SAMPLES, R[1], delta = EPS)

  def test_insomnia(self):
    P = pasp.parse("examples/insomnia.lp")
    A = ["insomnia", "work", "sleep"]
    R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
    S = pasp.sample(P, A, n = N_SAMPLES)

    # ℙ(insomnia)
    self.assertAlmostEqual(np.sum(S[:,0])/N_SAMPLES, R[0], delta = EPS)

    # ℙ(work)
    self.assertAlmostEqual(np.sum(S[:,1])/N_SAMPLES, R[1], delta = EPS)

    # ℙ(sleep)
    self.assertAlmostEqual(np.sum(S[:,2])/N_SAMPLES, R[2], delta = EPS)

  def test_multinsomnia(self):
    P = pasp.parse("examples/multinsomnia.lp")
    A = ["insomnia(anna)", "insomnia(bill)", "insomnia(charlie)",
         "work(anna)", "work(bill)", "work(charlie)",
         "sleep(anna)", "sleep(bill)", "sleep(charlie)",
         "calls(anna, bill)", "calls(anna, charlie)", "calls(bill, anna)", "calls(bill, charlie)",
         "calls(charlie, anna)", "calls(charlie, bill)"]
    R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
    S = np.sum(pasp.sample(P, A, n = N_SAMPLES), axis = 0)/N_SAMPLES

    self.assertTrue(np.allclose(S, R, atol = EPS))

  def test_smokers(self):
    P = pasp.parse("examples/smokers.lp")
    A = ["smokes(a)", "smokes(b)"]
    R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
    S = pasp.sample(P, A, n = N_SAMPLES)
    Q = np.sum(S, axis = 0) / N_SAMPLES

    # ℙ(smokes(a)), ℙ(smokes(b))
    self.assertTrue(np.allclose(Q, R, atol = EPS))

  def test_earthquake_ad(self):
    P = pasp.parse("examples/earthquake_ad.lp")
    A = ["alarm", "burglary", "earthquake(mild)", "earthquake(none)"]
    R = pasp.exact(P, quiet = True, psemantics = "maxent").flatten()
    S = pasp.sample(P, A, n = N_SAMPLES)

    # The probability of earthquake(heavy) is too low for the approximation of
    #   ℙ(alarm | burglary, earthquake(heavy)).
    # We shall skip this one.

    # ℙ(alarm | ¬burglary, earthquake(mild))
    p = np.sum(S[:,0] & ~S[:,1] & S[:,2]) # ℙ(alarm, ¬burglary, earthquake(mild))
    q = np.sum(~S[:,1] & S[:,2])          # ℙ(¬burglary, earthquake(mild))
    self.assertAlmostEqual(p/q, R[1], delta = EPS)

    # ℙ(alarm | burglary, ¬earthquake(mild))
    p = np.sum(S[:,0] & S[:,1] & ~S[:,2]) # ℙ(alarm, burglary, ¬earthquake(mild))
    q = np.sum(S[:,1] & ~S[:,2])          # ℙ(burglary, ¬earthquake(mild))
    self.assertAlmostEqual(p/q, R[2], delta = EPS)

    # ℙ(alarm | ¬burglary, earthquake(none))
    p = np.sum(S[:,0] & ~S[:,1] & S[:,3]) # ℙ(alarm, ¬burglary, earthquake(none))
    q = np.sum(~S[:,1] & S[:,3])          # ℙ(¬burglary, earthquake(none))
    self.assertAlmostEqual(p/q, R[3], delta = EPS)

if __name__ == "__main__":
  unittest.main()



import unittest
from .utils import PaspTest
import pasp
import numpy as np

class TestCounting(PaspTest):
  @staticmethod
  def all_learnable(P: pasp.program.Program):
    for pf in P.PF: pf.learnable = True
    for ad in P.AD: ad.learnable = True

  @staticmethod
  def init_example(eg: str, semantics = "stable"):
    P = pasp.parse("examples/" + eg + ".lp", semantics = semantics)
    TestCounting.all_learnable(P)
    return P, pasp.count(P)

  def assert_strat(self, P: pasp.program.Program, C: tuple):
    k = 2**len(P.PF)
    for ad in P.AD: k *= len(ad.P)
    self.assertTrue(np.all(C[0] == k//2))
    self.assertTrue(np.array_equal([i for i in range(len(P.PF))], C[1]))
    if len(P.AD) == 0:
      self.assertIsNone(C[2])
      self.assertIsNone(C[3])
    else:
      for i, ad in enumerate(P.AD):
        self.assertTrue(np.all(C[2][i] == k//len(ad.P)))
        self.assertEqual(C[3][i], i)

  def test_asia(self):
    P, C = self.init_example("asia")
    self.assert_strat(P, C)

  def test_earthquake(self):
    P, C = self.init_example("earthquake")
    self.assert_strat(P, C)

  def test_simple(self):
    P, C = self.init_example("simple")
    self.assert_strat(P, C)

  def test_simpler(self):
    P, C = self.init_example("simpler")
    self.assert_strat(P, C)

  def test_smokers(self):
    P, C = self.init_example("smokers")
    self.assert_strat(P, C)

  def test_earthquake_ad(self):
    P, C = self.init_example("earthquake_ad")
    self.assert_strat(P, C)

  def test_multinsomnia(self):
    P, C = self.init_example("multinsomnia")
    self.assertTrue(np.array_equal(C[0], [[54, 27], [54, 27], [54, 27]]))
    self.assertTrue(np.array_equal(C[1], [0, 1, 2]))
    self.assertTrue(np.array_equal(C[2][0], [27, 27, 27]))
    self.assertTrue(np.array_equal(C[3], [0]))

  def test_barber(self):
    P, C = self.init_example("barber", semantics = "lstable")
    self.assertTrue(np.array_equal(C[0], [[1, 1]]))
    self.assertTrue(np.array_equal(C[1], [0]))
    self.assertIsNone(C[2])
    self.assertIsNone(C[3])

  def test_game(self):
    P, C = self.init_example("game")
    self.assertTrue(np.array_equal(C[0], [[1, 2]]))
    self.assertTrue(np.array_equal(C[1], [0]))
    self.assertIsNone(C[2])
    self.assertIsNone(C[3])

  def test_insomnia(self):
    P, C = self.init_example("insomnia")
    self.assertTrue(np.array_equal(C[0], [[2, 1]]))
    self.assertTrue(np.array_equal(C[1], [0]))
    self.assertIsNone(C[2])
    self.assertIsNone(C[3])

if __name__ == "__main__":
  unittest.main()



import pathlib, enum, math, collections.abc
import lark, lark.reconstruct, clingo, numpy
from numpy import ascontiguousarray as contiguous
from .program import ProbFact, Query, ProbRule, Program, CredalFact, unique_fact, Semantics, Data
from .program import AnnotatedDisjunction, NeuralRule, NeuralAD

def is_fact(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a fact."
  return isinstance(x, lark.Tree) and x.data == "fact"
def is_pfact(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a probabilistic fact."
  return isinstance(x, lark.Tree) and x.data == "pfact"
def is_rule(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a rule."
  return isinstance(x, lark.Tree) and x.data == "rule"
def is_prule(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a probabilistic rule."
  return isinstance(x, lark.Tree) and x.data == "prule"
def is_atom(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a(n) (grounded) atom."
  return isinstance(x, lark.Tree) and (x.data == "atom" or x.data == "gratom")
def is_pred(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a (grounded) predicate."
  return isinstance(x, lark.Tree) and (x.data == "pred" or x.data == "grpred")

def atom_sign(x: lark.Tree) -> bool:
  "Returns whether a(n) (grounded) atom is negative (`False`), positive (`True`), or not an atom (`None`)."
  return x.children[0].type != "NEG" if is_atom(x) else None
def pred_sign(x: lark.Tree) -> bool:
  "Returns whether a (grounded) predicate is negative (`False`), positive (`True`) or not a predicate (`None`)."
  return x.children[0].type != "NEG" if is_pred(x) else None

def is_prob(x: lark.Tree) -> iter:
  "Returns whether a node (in this case a fact or rule) is probabilistic."
  return isinstance(x, lark.Tree) and len(x.children) > 0 and isinstance(x.children[0], lark.Token) and x.children[0].type == "prob"

def facts(x: lark.Tree) -> iter:
  "Returns an iterator containing all facts in the AST."
  return x.find_pred(is_fact)
def pfacts(x: lark.Tree) -> iter:
  "Returns an iterator containing all probabilistic facts in the AST."
  return x.find_pred(is_pfact)
def rules(x: lark.Tree) -> iter:
  "Returns an iterator containing all rules in the AST."
  return x.find_pred(is_rule)
def prules(x: lark.Tree) -> iter:
  "Returns an iterator containing all probabilistic rules in the AST."
  return x.find_pred(is_prule)
def probs(x: lark.Tree) -> iter:
  "Returns an iterator containing all probabilistic facts and rules in the AST."
  return x.find_pred(lambda x: is_prule(x) or is_pfact(x))

def expand_interval(x: lark.Tree) -> tuple[int, int]:
  assert isinstance(x, lark.Tree) and x.data == "interval", "Given AST node is not an Interval."
  return int(x.children[0].value), int(x.children[1].value)

def tree_contains(x: lark.Tree, f) -> bool:
  "Performs a depth-first search, returning (and stopping the search) node `y` when `f(y) == True`."
  V = set()
  def visit(x: lark.Tree | lark.Token) -> bool:
    V.add(x)
    if f(x): return True
    if isinstance(x, lark.Tree):
      for c in x.children:
        if (c not in V) and visit(c): return True
    return False
  return visit(x)

def is_var(x: lark.Token): return isinstance(x, lark.Token) and x.type == "VAR"

def is_nonground(x: lark.Tree) -> bool:
  "Returns whether node `x` is not grounded, i.e. whether any node in the subtree of `x` is a variable."
  try: iter(x)
  except TypeError: return tree_contains(x, is_var)
  else: return any(tree_contains(x, is_var))
def is_ground(x: lark.Tree) -> bool:
  "Returns whether node `x` is grounded, i.e. whether no node in the subtree of `x` is a variable."
  return not is_nonground(x)

def read(*files: str, G: lark.Lark = None, from_str: bool = False) -> lark.Tree:
  "Read all `files` and parse them with grammar `G`, returning a single `lark.Tree`."
  if G is None:
    try:
      with open(pathlib.Path(__file__).resolve().parent.joinpath("grammar.lark"), "r") as f:
        G = lark.Lark(f, start = "plp")
    except Exception as ex:
      raise ex
  T = None
  if from_str: T = G.parse("\n".join(files)); return T
  for fname in files:
    try:
      # For now, dump files entirely into memory (this isn't too much of a problem, since PLPs are
      # usually small). In the future, consider streaming batches of text instead for large files.
      with open(fname, "r") as f:
        if T is None: T = G.parse(f.read())
        else:
          U = G.parse(f.read())
          T.children.extend(u for u in U.children if u not in T.children)
    except Exception as ex:
      raise ex
  assert T is not None, "No file read."
  return T

def getnths(X: iter, i: int) -> iter: return (x[i] for x in X)
def find(X: iter, v, d = None):
  try:
    i = X.index(v)
    return i, X[i]
  except ValueError: return -1, d
def push(L: list, X: iter):
  "Pushes X to L. If X is a list, then push all elements of X to L instead of the list object itself."
  if isinstance(X, list): L.extend(X)
  else: L.append(X)

def lit2atom(x: str) -> str: return x[4:] if x[:4] == "not " else x

class StableTransformer(lark.Transformer):
  def __init__(self, _):
    super().__init__()
    self.semantics = Semantics.STABLE
    self.torch_scope = {}

  @staticmethod
  def pack(t: str, rep: str = None, val = None, scope: dict = {}) -> tuple[str, str, str, dict]:
    class Pack(tuple):
      @staticmethod
      def __new__(cls, tp: str, r: str = None, v = None, sc: dict = {}):
        return super(Pack, cls).__new__(cls, (tp, str(v) if r is None else r, r if v is None else v, sc))
      def __str__(self): return self[1]
      def __repr__(self): return self.str()
    return Pack(t, rep, val, scope)

  @staticmethod
  def join_scope(A: list) -> dict: return dict((y, None) for S in A for y in S[3])

  @staticmethod
  def find_data_pred(D: dict, body: list, which: str, name: str) -> list:
    t = None
    for d in body:
      if d[2][1] in D:
        t = D[d[2][1]]
        break
    if t is None: raise ValueError(f"Neural {which} {name} must contain a data predicate!")
    return t

  @staticmethod
  def check_data(D: list):
    "Checks if all data have same first dimension size."
    n, m = -1, -1
    for X in D.values():
      if n < 0: n = X[0].test.shape[0]
      if (X[0].train is not None) and (m < 0): m = X[0].train.shape[0]
      for x in X:
        if x.test.shape[0] != n:
          raise ValueError("Test data must have same number of instances!")
        if (x.train is not None) and (x.train.shape[0] != m):
          raise ValueError("Train data must have same number of instances!")

  @staticmethod
  def register_nrule(TNR: list, NR: list, D: list):
    for name, inp, net, body, rep, learnable, params in TNR:
      t = StableTransformer.find_data_pred(D, body, "rule", name)
      # Ground rules.
      H = contiguous(tuple(clingo.parse_term(f"{name}({t[i].arg})")._rep for i in range(len(t))), \
                      dtype = numpy.uint64)
      B, S = None, None
      if len(body) > 1:
        body_no_data = [b for b in body if b[2][1] != t[0].name]
        B = contiguous(tuple(clingo.parse_term(f"{b[2][1]}({t[i].arg})" if len(b[3]) > 0 \
                                                else lit2atom(b[1]))._rep for i in range(len(t)) \
                              for b in body_no_data), dtype = numpy.uint64)
        S = contiguous(tuple(b[2][0] for i in range(len(t)) for b in body_no_data), dtype = bool)
      NR.append(NeuralRule(H, B, S, name, net, rep, t, learnable, params))

  @staticmethod
  def register_nad(TNA: list, NA: list, D: list):
    for name, inp, vals, net, body, rep, learnable, params in TNA:
      t = StableTransformer.find_data_pred(D, body, "AD", name)
      # Ground rules.
      V = list(vals.keys())
      H = contiguous(tuple(clingo.parse_term(f"{name}({t[i].arg}, {V[j]})")._rep \
                            for i in range(len(t)) for j in range(len(V))))
      B, S = None, None
      if len(body) > 1:
        body_no_data = [b for b in body if b[2][1] != t[0].name]
        # B and S do not depend on the number of values |V|, only on |t| and |body|.
        B = contiguous(tuple(clingo.parse_term(f"{b[2][1]}({t[i].arg})" if len(b[3]) > 0 \
                                                else lit2atom(b[1]))._rep for i in range(len(t)) \
                              for b in body_no_data), dtype = numpy.uint64)
        S = contiguous(tuple(b[2][0] for i in range(len(t)) for b in body_no_data), dtype = bool)
      NA.append(NeuralAD(H, B, S, name, V, net, rep, t, learnable, params))

  # Components which are directly translated to clingo.
  def CMP_OP(self, o): return self.pack("CMP_OP", str(o))
  def aggr(self, A): return self.pack("aggr", "".join(str(x) for x in A))

  # Terminals.
  def UND(self, u): return self.pack("UND", str(u))
  def WORD(self, c): return self.pack("WORD", str(c))
  def NEG(self, n): return self.pack("NEG", str(n))
  def VAR(self, v):
    x = str(v); X = {v: None}
    return self.pack("VAR", x, scope = X)
  def ID(self, i): return self.pack("ID", val = int(i))
  def OP(self, o): return self.pack("OP", str(o))
  def REAL(self, r): return self.pack("REAL", val = float(r))
  def frac(self, f): return self.pack("frac", val = f[0][2]/f[1][2])
  def prob(self, p): return self.pack("prob", val = p[0][2])
  def EXPAND(self, e): return self.pack("EXPAND",  str(e))
  def LEARN(self, l): return self.pack("LEARN", str(l))
  def CONST(self, c): return self.pack("CONST", str(c))
  def BOOL(self, b): return self.pack("BOOL", v := b.lower(), v != "false")
  def NULL(self, n): return self.pack("NULL", None, None)

  # Path.
  def path(self, p): return self.pack(p[0].type, p[0].value)

  # Set.
  def set(self, S):
    M = dict((x[1], None) for x in S)
    if len(M) != len(S): raise ValueError("set must contain only unique constants!")
    return self.pack("set", f"{{{','.join(x for x in M.keys())}}}", M)

  # Intervals.
  def interval(self, I): return self.pack("interval", f"{I[0][2]}..{I[1][2]}")

  # Predicates.
  def pred(self, P):
    name = P[0][1]
    return self.pack("pred", f"{name}({', '.join(getnths(P[1:], 1))})", name, self.join_scope(P))
  def grpred(self, P): return self.pred(P)

  # Literals.
  def lit(self, P):
    s = P[0][0] != "NEG"
    return self.pack("lit", " ".join(getnths(P, 1)), (s, P[0][2] if s else P[1][2]), self.join_scope(P))
  def grlit(self, P): return self.lit(P)

  # Binary operations.
  def bop(self, B) -> str: return self.pack("bop", " ".join(getnths(B, 1)))

  # Facts.
  def fact(self, F):
    f = f"{''.join(getnths(F, 1))}"
    # Facts are always grounded.
    return self.pack("fact", f + ".", f)
  def pfact(self, PF):
    p, f = PF[0][2], PF[1][1]
    return self.pack("pfact", "", ProbFact(p, f))
  def cfact(self, CF):
    l, u, f = CF[0][2], CF[1][2], CF[2][1]
    return self.pack("cfact", "", CredalFact(l, u, f))
  def lpfact(self, PF):
    if PF[0][0] == "prob": p, f = PF[0][2], PF[1][1]
    else: p, f = 0.5, PF[0][2]
    return self.pack("pfact", "", ProbFact(p, f, learnable = True))

  # Heads.
  def head(self, H): return self.pack("head", ", ".join(getnths(H, 1)), H, self.join_scope(H))
  def ohead(self, H): return self.pack("head", H[0][1], H[0][2], H[0][3])
  # Body.
  def body(self, B): return self.pack("body", ", ".join(getnths(B, 1)), B, self.join_scope(B))

  # Rules.
  def rule(self, R): return self.pack("rule", " :- ".join(getnths(R, 1)) + ".")
  def prule(self, R):
    l = "LEARN" in getnths(R, 0)
    e = "EXPAND" in getnths(R, 0)
    h, b = R[-2], R[-1]
    o = f"{h[1]} :- {b[1]}"
    p = R[0][2]
    if e:
      S = self.join_scope(R)
      if len(S) == 0:
        pr = ProbRule(p, o, is_prop = True)
        return self.pack("prule", pr.prop_f, pr)
      # Invariant: len(b) > 0, otherwise the rule is unsafe.
      name = h[2]
      # hscope is guaranteed to be ordered by Python dict's definition.
      hscope = h[3]
      body_preds = [x for x in b[2] if x[0] != "bop"]
      h_s = ", ".join(hscope) + ", " if len(hscope) > 0 else ""
      b_s = ", ".join(map(lambda x: f"1, {x[1]}" if x[2][0] else f"0, {x[1][4:]}", body_preds))
      # The number of body arguments is twice as we need to store the sugoal's sign and symbol.
      u = f"{name}(@unify(\"{p}\", {name}, {int(l)}, {len(hscope)}, {2*len(body_preds)}, {h_s}{b_s})) :- {b[1]}."
      return self.pack("prule", "", ProbRule(p, o, is_prop = False, unify = u, learnable = l))
    else:
      # TODO: parameter tying
      pr = ProbRule(p, o, is_prop = True, learnable = l)
      return self.pack("prule", pr.prop_f, pr)

  # Annotated disjunction head.
  def ad_head(self, H):
    P, F = [], []
    for i in range(0, len(H), 2):
      P.append(H[i][2])
      F.append(H[i+1][1])
    return self.pack("ad_head", F, P, self.join_scope(H))
  # Learnable annotated disjunction head.
  def lad_head(self, H: list):
    P, F = [], []
    i, o, j = 0, 0, 0
    last = None
    while i < len(H):
      a = H[i]
      if a[0] == "prob":
        P.append(a[2])
        F.append(H[i+1][1])
        i += 2
      else:
        P.append(-1)
        F.append(a[1])
        i += 1; o += 1
        last = j
      j += 1
    if o > 0:
      P_s = sum(P)+o
      # If probs were not explicitly given, assume maximum uncertainty and set to uniform.
      s = round((1.0-P_s)/o, ndigits = 15)
      ts = P_s+s*(o-1)
      for i, p in enumerate(P):
        if i == last: P[i] = 1.0-ts
        elif p < 0: P[i] = s
    return self.pack("lad_head", F, P, self.join_scope(H))
  # Annotated disjunctions.
  def ad(self, AD):
    P, F, learnable = AD[0][2], AD[0][1], AD[0][0] == "lad_head"
    if not math.isclose(s := sum(P), 1.0):
      P.append(1-s)
      F.append(unique_fact())
    return self.pack("ad", "", AnnotatedDisjunction(P, F, learnable), AD[0][3])
  def adr(self, AD):
    raise NotImplementedError

  def _data2tensor(self, D):
    tp, path_or_func = D[0][0], D[0][2]
    import pandas, numpy
    # Is an external file or URL.
    if tp != "PY_FUNC": data = pandas.read_csv(path_or_func, dtype = numpy.float32)
    else: # is a function.
      if path_or_func not in self.torch_scope:
        raise ValueError(f"No data definition {path_or_func} found! Either define it in a PyTorch "
                         "block or specify a file or URL to read from.")
      data = self.torch_scope[path_or_func]()
      try:
        import torch
      except ModuleNotFoundError:
        raise ModuleNotFoundError("PyTorch not found! PyTorch must be installed for neural rules "
                                  "and neural ADs.")
      if not issubclass(type(data), torch.Tensor): data = torch.tensor(data)
    return data

  # Test set special predicate.
  def test(self, T): return self.pack("test", "", self._data2tensor(T))
  # Train set special predicate.
  def train(self, R): return self.pack("train", "", self._data2tensor(R))

  # Data special predicate.
  def data(self, D):
    name, arg = D[0][1], D[1][1]
    test, train = D[2][2], D[3][2] if len(D) > 3 else None
    return self.pack("data", f"{name}({arg}).", Data(name, arg, test, train))

  # Torch block.
  def torch(self, T):
    exec("import torch\n\n" + T[0].value, self.torch_scope)
    return self.pack("torch", "")

  # Local hubconf repo.
  def LOCAL_NET(self, L): return self.pack("LOCAL_NET", str(L))
  # GitHub hubconf repo.
  def GITHUB(self, H): return self.pack("GITHUB", str(H))
  # Python function.
  def PY_FUNC(self, P): return self.pack("PY_FUNC", str(P))

  # Hub network.
  def hub(self, H):
    # Function name or entrypoint.
    func = H[0][1]
    # Network is coming from a Torch block.
    if len(H) == 1:
      if func not in self.torch_scope:
        raise ValueError(f"No network definition {func} found! Either define it in a PyTorch "
                         "block or specify a PyTorch Hub model (local or from GitHub).")
      N = self.torch_scope[func]()
      rep = f"@{func}"
    # Network is coming from PyTorch Hub.
    else:
      try:
        import torch
      except ModuleNotFoundError:
        raise ModuleNotFoundError("PyTorch not found! PyTorch must be installed for neural rules "
                                  "and neural ADs.")
      path, source = H[1][1], "github" if H[1][0] == "GITHUB" else "local"
      N = torch.hub.load(path, func, source = source, trust_repo = "check")
      rep = f"@{func} on \"{path}\" at \"{source}\""
    return self.pack("hub", "", (N, rep))

  # Optimizer parameters.
  def params(self, P):
    return self.pack("params", "", {P[i][1]: P[i+1][2] for i in range(0, len(P), 2)})

  # Neural rule.
  def nrule(self, A):
    learnable = A[0][0] == "LEARN"
    name = A[1][1]
    inp = A[2][1]
    net, hub_repr = A[3][2]
    if A[4][0] == "params":
      params = A[4][2]
      body = A[5:]
    else:
      params = {}
      body = A[4:]
    scope = self.join_scope(A)

    if len(scope) != 1:  raise ValueError(f"Neural rule {name} is not grounded!")
    if inp not in scope: raise ValueError(f"Neural rule {name} is unsafe!")

    rep = f"{A[0][1]}::{name}({inp}) as {hub_repr} :- {', '.join(getnths(body, 1))}."
    return self.pack("nrule", "", (name, inp, net, body, rep, learnable, params))

  # Neural annotated disjunction.
  def nad(self, A):
    learnable = A[0][0] == "LEARN"
    name = A[1][1]
    inp = A[2][1]
    vals = A[3][2]
    net, hub_repr = A[4][2]
    if A[5][0] == "params":
      params = A[5][2]
      body = A[6:]
    else:
      params = {}
      body = A[5:]
    scope = self.join_scope(A)

    if len(scope) != 1:  raise ValueError(f"Neural annotated disjunction {name} is not grounded!")
    if inp not in scope: raise ValueError(f"Neural annotated disjunction {name} is unsafe!")

    rep = f"{A[0][1]}::{name}({inp}, {A[2][1]}) as {hub_repr} :- {', '.join(getnths(body, 1))}."
    return self.pack("nad", "", (name, inp, vals, net, body, rep, learnable, params))

  # Constraint.
  def constraint(self, C): return self.pack("constraint", f":- {C[0][1]}.")

  # Query elements.
  def qelement(self, E): return self.pack("qelement", " ".join(getnths(E, 1)))
  # Interpretations.
  def interp(self, I): return self.pack("interp", "", getnths(I, 1))
  # Queries.
  def query(self, Q):
    return self.pack("query", "", Query(Q[0][2], Q[1][2] if len(Q) > 1 else [], semantics = self.semantics))

  # Constant definition.
  def constdef(self, C): return self.pack("constdef", f"#const {C[0][1]} = {C[1][1]}.")

  # Probabilistic Logic Program.
  def plp(self, C) -> Program:
    # Logic Program.
    P  = []
    # Probabilistic Facts.
    PF = []
    # Probabilistic Rules.
    PR = []
    # Queries.
    Q  = []
    # Credal Facts.
    CF = []
    # Annotated Disjunction.
    AD = []
    # Neural arguments and data.
    TNR, TNA = [], []
    D = {}
    # Actual neural rules and neural ADs.
    NR, NA = [], []
    # Mapping.
    M = {"pfact": PF, "prule": PR, "query": Q, "cfact": CF, "ad": AD, "nrule": TNR, "nad": TNA}
    for t, L, O, _ in C:
      if len(L) > 0: push(P, L)
      if t in M: push(M[t], O)
      if t == "data":
        if O.name in D: D[O.name].append(O)
        else: D[O.name] = [O]
    # Deal with ungrounded probabilistic rules.
    for r in PR:
      if r.is_prop: PF.append(r.prop_pf)
    self.check_data(D)
    self.register_nrule(TNR, NR, D)
    self.register_nad(TNA, NA, D)
    return Program("\n".join(P), PF, PR, Q, CF, AD, NR, NA, semantics = self.semantics)

class PartialTransformer(StableTransformer):
  def __init__(self, sem: str):
    super().__init__(sem)
    self.PT = set()
    self.semantics = Semantics.LSTABLE if sem == "lstable" else Semantics.PARTIAL
    self.o_tree = None

  @staticmethod
  def has_binop(x: str): return ("=" in x) or ("<" in x) or (">" in x)

  def fact(self, F):
    T = super().fact(F)
    self.PT.add(T[2])
    return T

  def pfact(self, PF):
    T = super().pfact(PF)
    self.PT.add(T[2].f)
    return T

  def cfact(self, CF):
    T = super().cfact(CF)
    self.PT.add(T[2].f)
    return T

  def rule(self, R):
    b1 = ", ".join(map(lambda x: x[1] if x[2][0] else f"not _{x[1][4:]}", R[1][2]))
    b2 = ", ".join(map(lambda x: f"_{x[1]}" if x[2][0] or PartialTransformer.has_binop(x) else x[1], R[1][2]))
    h1, h2 = R[0][1], ", ".join(map(lambda x: f"_{x[1]}", R[0][2]))
    for h in R[0][2]: self.PT.add(h[1])
    # for x in r[1][3]:
      # if not PartialTransformer.has_binop(x): self.PT.add(x[4:] if x[:4] == "not " else x)
    return self.pack("rule", [f"{h1} :- {b1}.", f"{h2} :- {b2}."])

  def prule(self, R):
    l = "LEARN" in getnths(R, 0)
    e = "EXPAND" in getnths(R, 0)
    p = R[0][2]
    h, b = R[-2], R[-1]
    tr_negs = lambda x: x[1] if x[2][0] else f"not _{x[1][4:]}"
    tr_pos  = lambda x: f"_{x[1]}" if x[2][0] or PartialTransformer.has_binop(x) else x[1]
    b1 = ", ".join(map(tr_negs, b[2]))
    b2 = ", ".join(map(tr_pos, b[2]))
    o1, o2 = f"{h[1]} :- {b1}", f"_{h[1]} :- {b2}"
    self.PT.add(h[1])
    uid = unique_fact()
    if e:
      S = self.join_scope(R)
      if len(S) == 0:
        pr1, pr2 = ProbRule(p, o1, ufact = uid), ProbRule(p, o2, ufact = uid)
        return self.pack("prule", [pr1.prop_f, pr2.prop_f], [pr1, pr2])
      # Invariant: len(b) > 0, otherwise the rule is unsafe.
      name = h[2]
      hscope = h[3]
      body_preds = [x for x in b[2] if x[0] != "bop"]
      h_s = ", ".join(hscope) + ", " if len(hscope) > 0 else ""
      b1_s = ", ".join(map(lambda x: f"1, {x[1]}" if x[2][0] else f"0, _{x[1][4:]}", body_preds))
      # Let the grounder deal with the _f rule.
      u1 = f"{name}(@unify(\"{p}\", {name}, {int(l)}, {len(hscope)}, {2*len(body_preds)}, {h_s}{b1_s})) :- {b1}."
      return self.pack("prule", "", ProbRule(p, o1, is_prop = False, unify = u1, learnable = l))
    else:
      pr1, pr2 = ProbRule(p, o1, ufact = uid, learnable = l), ProbRule(p, o2, ufact = uid)
      return self.pack("prule", [pr1.prop_f, pr2.prop_f], [pr1, pr2])

  def plp(self, C: list[tuple]) -> Program:
    # Logic Program.
    P  = []
    # Probabilistic Facts.
    PF = []
    # Probabilistic Rules.
    PR = []
    # Queries.
    Q  = []
    # Credal Facts.
    CF = []
    # Annotated Disjunction.
    AD = []
    # Neural rules and ADs.
    NR, NA = [], []
    # Mapping.
    M = {"pfact": PF, "prule": PR, "query": Q, "cfact": CF, "ad": AD}
    for t, L, O, _ in C:
      if len(L) > 0: push(P, L)
      if t in M: push(M[t], O)
      if t == "prule" and isinstance(O, collections.abc.Iterable) and O[0].is_prop:
        PF.append(O[0].prop_pf)
    P.extend(f"_{x} :- {x}." for x in self.PT)
    return Program("\n".join(P), PF, PR, Q, CF, AD, NR, NA, semantics = self.semantics, \
                   stable_p = self.stable_p)

  def transform(self, tree):
    self.o_tree = tree
    self.stable_p = StableTransformer(self.semantics).transform(tree)
    return super().transform(tree)

def parse(*files: str, G: lark.Lark = None, from_str: bool = False, semantics: str = "stable") -> Program:
  """Either parses `streams` as blocks of text containing the PLP when `from_str = True`, or
  interprets `streams` as filenames to be read and parsed into a `Program`."""
  if semantics not in parse.trans_map:
    raise ValueError("semantics not supported (must either be 'stable', 'partial' or 'lstable')!")
  return parse.trans_map[semantics](semantics).transform(read(*files, G = G, from_str = from_str))
parse.trans_map = {}
parse.trans_map["stable"] = StableTransformer
parse.trans_map["lstable"] = PartialTransformer
parse.trans_map["partial"] = PartialTransformer




import enum

import clingo
from clingo.symbol import Function

def unique_fact(i: int = None) -> str:
  """
  Creates a new unique fact for probabilistic rules. To do this, we update a counter `unique_fact.i`
  in a way equivalent to C's `static` variables.
  """
  if i is None:
    unique_fact.i += 1
    return f"__unique_id_{unique_fact.i}"
  return f"__unique_id_{i}"
unique_fact.i = 0

class ProbFact:
  """
  A Probabilistic Fact (PF) is a (Logic Program) fact which is "chosen" with some probability.
  """

  def __init__(self, p: str, f: str, learnable: bool = False):
    "Constructs a PF out of a probability `p` and fact `f`."
    self.p = float(p)
    self.f = f
    # Construct a clingo.symbol.Function from this fact.
    self.cl_f = clingo.parse_term(f)
    self.learnable = learnable

  def __str__(self) -> str: return f"{round(self.p, ndigits = 3)}{'?' if self.learnable else ''}::{self.f}"
  def __repr__(self) -> str: return self.__str__()

class ProbRule:
  """
  A Probabilistic Rule (PR) is a (Logic Program) rule that (when propositional) may be chosen with
  some probability `p`. A non-propositional PR must be grounded first.
  """

  def __init__(self, p: str, f: str, is_prop: bool = True, unify: str = None, ufact: str = None,
               learnable: bool = False):
    self.p = p
    self.f = f
    self.is_prop = is_prop
    self.learnable = learnable
    self.unify = unify
    self.prop_pf = ProbFact(p, unique_fact() if ufact is None else ufact, learnable = learnable)
    self.prop_f = f"{f}, {self.prop_pf.f}."

  def __str__(self) -> str: return f"{self.p}{'+' if self.unify else '' + '?' if self.learnable else ''}::{self.f}"
  def __repr__(self) -> str: return self.__str__()

class CredalFact:
  """
  A Credal Fact (CF) consists of a fact `f` attached to a probability interval `[l, u]`, where `l ∈
  [0, 1]` is the lowest probability `f` may attain and `u ≥ l ∈ [0, 1]` is the highest.
  """

  def __init__(self, l: float, u: float, f: str):
    self.l, self.u = float(l), float(u)
    self.f = f
    self.cl_f = clingo.parse_term(f)

  def __getitem__(self, i: bool) -> float: return self.l if False else self.u
  def __str__(self) -> str: return f"[{self.l}, {self.u}]::{self.f}"
  def __repr__(self) -> str: return self.__str__()

def _str_query_assignment(f: Function, t: bool) -> str:
  """
  String formats a query tuple `(f, t)`, where `f` is an atom and `t` is whether it should appear
  in the program or not.
  """
  return str(f) if t == Query.TERM_POS else ("not " + str(f) if t == Query.TERM_NEG else "undef " + str(f))

class AnnotatedDisjunction:
  def __init__(self, P: list[float], F: list[str], learnable: bool = False):
    self.P = P
    self.F = F
    self.cl_F = [clingo.parse_term(f) for f in F]
    self.learnable = learnable

  def __getitem__(self, i: int) -> tuple[float, str]:
    return self.P[i], self.F[i]
  def __str__(self) -> str:
    return "; ".join([f"{(p if (p := round(self.P[i], ndigits = 3)) > 0 else '*')}{'?' if self.learnable else ''}::{self.F[i]}" for i in range(len(self.P))])
  def __repr__(self) -> str: return self.__str__()

try:
  import torch
except ModuleNotFoundError:
  print("PyTorch not found! PyTorch must be installed for neural rules and neural ADs to be used "
        "in programs.")

class Data:
  def __init__(self, name: str, arg: str, test, train = None):
    self.name = name
    self.arg = arg
    import pandas
    if issubclass(type(test), pandas.DataFrame): self.test = torch.tensor(test.to_numpy())
    else: self.test = test
    if issubclass(type(train), pandas.DataFrame): self.train = torch.tensor(train.to_numpy())
    else: self.train = train

    if self.train is not None:
      assert self.test.shape[1:] == self.train.shape[1:], \
        "Train and test sets must have same shape (excluding the first dimension)!"

  def __str__(self):
    if self.train is None: return f"{self.name}({self.arg}) ~ test({self.test.shape})"
    else: return f"{self.name}({self.arg}) ~ test({self.test.shape}), train({self.train.shape})"
  def __repr__(self): return self.__str__()

class Neural:
  def __init__(self, net, data: Data, learnable: bool, rep: str, nvals: int,
               opt_params: dict):
    self.net = net
    self.learnable = learnable
    self.rep = rep
    self.data = data

    # Update default opt_params with given params.
    _opt_params = {"lr": 0.01, "maximize": True}
    _opt_params.update(opt_params)
    self.opt = torch.optim.SGD(net.parameters(), **_opt_params)

    self.test = torch.cat(tuple(d.test for d in data), dim = 0)
    self.train = None if data[0].train is None else torch.cat(tuple(d.train for d in data), dim = 0)
    self.out = None
    # Derivatives of the logic program to be passed to backwards.
    self.dw = torch.zeros((self.train.shape[0], nvals)) if learnable else None
    self.net.train()

  def __str__(self): return self.rep
  def __repr__(self): return self.__str__()

  def pr(self):
    "Retrieves the probabilities of the neural rule from the test set."
    with torch.inference_mode():
      return self.net(self.test).cpu().numpy()

  def forward(self, start: int = 0, end: int = None):
    "Retrieves the probabilities of the neural rule from the train set."
    if self.learnable:
      self.out = self.net(self.train[start:end])
      return self.out.data.cpu().numpy()
    with torch.inference_mode():
      return self.net(self.train[start:end]).cpu().numpy()

  def backward(self, start: int = 0, end: int = None):
    """ Performs backpropagation and runs the optimizer step.
    Argument `dl` is the derivative of the program as a `numpy.ndarray`.
    """
    self.out[start:end].backward(self.dw[start:end])
    self.opt.step()
    self.opt.zero_grad()

  def ntest(self): return self.data[0].test.shape[0]
  def ntrain(self): return self.data[0].train.shape[0] if self.learnable else 0

class NeuralRule(Neural):
  def __init__(self, heads: list, bodies: list, signs: list, name: str, net, rep: str, data: list,
               learnable: bool, params: dict):
    super().__init__(net, data, learnable, rep, 1, params)
    # Heads and bodies must be numpy.uint64 values representing _rep, not Symbols.
    self.H = heads
    self.B = bodies
    self.S = signs
    self.name = name

    # Validate net during parsing so that it won't blow up in our faces during inference or learning.
    p = self.pr()
    assert p.ndim == 2, \
           "Networks embedded onto neural rules must output a single probability!"

class NeuralAD(Neural):
  def __init__(self, heads: list, bodies: list, signs: list, name: str, vals: list, net, rep: str, \
               data: list, learnable: bool, params: dict):
    super().__init__(net, data, learnable, rep, len(vals), params)
    self.H = heads
    self.B = bodies
    self.S = signs
    self.name  = name
    self.vals  = vals

    # Validate net during parsing so that it won't blow up in our faces during inference or learning.
    p = self.pr()
    assert p.ndim == 2, \
           "Networks embedded onto neural rules must output a 1D probability tensor!"

class Semantics(enum.IntEnum):
  STABLE = 0
  PARTIAL = 1
  LSTABLE = 2

class Query:
  """
  A query is a meta-command within a PLP to signal the solver to produce and output a probabilistic
  query. A query follows a modified PASOCS [1] syntax, that is, the query (not necessarily in this
  order)

  ```
  #query(q1; ...; qk; not p1; ...; not pm | e1; ...; en; not v1; ...; not vt)
  ```

  of ground atoms `q1, ..., qk`, `p1, ..., pm`, `e1, ..., en`, `v1, ..., vt` is equivalent to asking
  the probability

  ```
  P({q1, ..., qk} = true, {p1, ..., pm} = false | {e1, ..., en} = true, {v1, ..., vt} = false).
  ```

  See concrete examples in the `/examples` folder.

  [1] - PASOCS: A Parallel Approximate Solver for Probabilistic Logic Programs under the Credal
  Semantics. Tuckey et al, 2021. URL: https://arxiv.org/abs/2105.10908.
  """

  TERM_NEG = 0
  TERM_POS = 1
  TERM_UND = 2

  def __init__(self, Q: iter, E: iter = [], semantics: Semantics = Semantics.STABLE):
    """
    Constructs a query from query (`Q`) and evidence (`E`) assignments.

    We use the notation `iter` as a type hinting to mean `Q` and `E` are iterables.
    """
    self.Q = [Query.parse_term(q, semantics) for q in Q]
    self.E = [Query.parse_term(e, semantics) for e in E]

  @staticmethod
  def parse_term(u: str, s: Semantics):
    if u.startswith("not "): t, n = u[4:], Query.TERM_NEG
    elif u.startswith("undef "): t, n = u[6:], Query.TERM_UND
    else: t, n = u, Query.TERM_POS
    return clingo.parse_term(t), n, None if s == Semantics.STABLE else clingo.parse_term(f"_{t}")

  def __str__(self) -> str:
    qs = f"ℙ({', '.join(_str_query_assignment(q, t) for q, t, _ in self.Q)}"
    if len(self.E) != 0: return qs + f" | {', '.join(_str_query_assignment(e, t) for e, t, _ in self.E)})"
    return qs + ")"
  def __repr__(self) -> str: return self.__str__()

class Program:
  """
  A Probabilistic Logic Program (PLP) usually configures a triple `<P,PF,CF>`, where `P` is a logic
  program, `PF` are probabilistic facts and `CF` are credal facts. We extend a PLP into a tuple
  `<P,PF,CF,Q>`, where `Q` are the queries to be asked from `P`, `PF` and `CF`.

  We accept ProbLog's syntactic sugar for probabilistic rules,

  ```
  p::h(X) :- b1(X), b2(X), ..., bn(X).
  ```

  meaning that if `b1(X), b2(X), ..., bn(X)` is true, `h(X)` is added with probability `p`. This is
  equivalent to

  ```
  p::a.
  h(X) :- b1(X), b2(X), ..., bn(X), a.
  ```

  where `a` is a unique probabilistic fact added with probability `p`.
  """

  def __init__(self, P: str, PF: list[ProbFact], PR: list[ProbRule], Q: list[Query], \
               CF: list[CredalFact], AD: list[AnnotatedDisjunction], NR: list[NeuralRule], \
               NA: list[NeuralAD], semantics: Semantics = Semantics.STABLE, stable_p = None):
    """
    Constructs a PLP out of a logic program `P`, probabilistic facts `PF`, credal facts `CF` and
    queries `Q`.
    """
    self.P = P
    self.PF = PF
    self.PR = PR
    self.Q = Q
    self.CF = CF
    self.AD = AD
    self.NR = NR
    self.NA = NA

    # Number of instances in data.
    self.m_test = 0
    self.m_train = 0
    if (len(NR) > 0) or (len(NA) > 0):
      self.m_test  = NR[0].ntest() if len(NR) > 0 else NA[0].ntest()
      for nr in (NR + NA):
        if nr.learnable:
          self.m_train = nr.ntrain()
          break

    self.gr_P = ""

    self.semantics = semantics
    self.stable = stable_p

  @staticmethod
  def str_if_contains(s: str, L):
    return f"\n{s}:\n{L}," if len(L) > 0 else ""

  def __str__(self) -> str:
    return f"<Logic Program:\n{self.P}," + \
           self.str_if_contains("Probabilistic Facts", self.PF) + \
           self.str_if_contains("Credal Facts", self.CF) + \
           self.str_if_contains("Annotated Disjunctions", self.AD) + \
           self.str_if_contains("Probabilistic Rules", self.PR) + \
           self.str_if_contains("Neural Rules", self.NR) + \
           self.str_if_contains("Neural Annotated Disjunctions", self.NA) + \
           f"\nQueries:\n{self.Q}>"
  def __repr__(self) -> str: return self.__str__()



"""
.. include:: ../README.md
"""

from .grammar import parse
from exact import exact, count
from ground import ground
from .program import Program
from sample import sample

import numpy as np

__version__ = "0.0.2-5"

def learn(P: Program, D: np.ndarray, A: np.ndarray = None, niters: int = 30, alg: str = "fixpoint",
          eta: float = 0.001, batch: int = 100, lstable_sat: bool = True):
  if A is None:
    if type(D) is np.ndarray: data = D if np.issubdtype(A.dtype, bytes) else A.astype(bytes)
    else: data = D
    from learn import learn_batch as clearn_batch
    return clearn_batch(P, data, niters = niters, alg = alg, eta = eta, batch = batch, lstable_sat = lstable_sat)

  if type(A) is not np.ndarray: atoms = np.array(A, dtype = bytes)
  else: atoms = A if np.issubdtype(A.dtype, bytes) else A.astype(bytes)
  if type(D) is not np.ndarray: data = np.array(D, dtype = np.uint8)
  else: data = D if np.issubdtype(D.dtype, np.uint8) else D.astype(np.uint8)

  obs, obs_counts = np.unique(data, axis = 0, return_counts = True)
  from learn import learn as clearn
  clearn(P, obs, obs_counts, atoms, niters = niters, alg = alg, eta = eta, lstable_sat = lstable_sat)



import os
from setuptools import setup, Extension, find_packages, Command
import numpy as np

class TestCommand(Command):
  description = "Runs unit and functional tests for PASP."
  user_options = []

  def initialize_options(self):
    self.cwd = None

  def finalize_options(self):
    self.cwd = os.getcwd()

  def run(self):
    assert os.getcwd() == self.cwd, f"Must be in package root: {self.cwd}"
    os.system("python setup.py build_ext --inplace && " \
              "python -m unittest tests/examples.py tests/counting.py tests/sampling.py -b")

# Debug concurrency problems by forcing sequential running.
# STD_MACROS = [("NUM_PROCS", str(1)), ("_GNU_SOURCE", None)]
STD_MACROS = [("NUM_PROCS", str(os.cpu_count())), ("_GNU_SOURCE", None)]

exact    = Extension("exact",
                     libraries = ["m", "clingo", "pthread"],
                     depends = ["pasp/cprogram.c", "pasp/coptimize.c", "pasp/cinf.c",
                                "pasp/cutils.c", "pasp/carray.c", "pasp/ground.c",
                                "pasp/cexact.c"],
                     sources = ["pasp/exact.c", "thpool/thpool.c", "pasp/cinf.c",
                                "bitvector/bitvector.c", "pasp/cutils.c", "pasp/coptimize.c",
                                "pasp/carray.c", "pasp/cprogram.c", "pasp/cexact.c"],
                     include_dirs = [np.get_include()],
                     extra_compile_args = ["-Wno-unused-function"],
                     define_macros = STD_MACROS)
ground   = Extension("ground",
                     libraries = ["clingo"],
                     depends = ["pasp/cutils.c", "pasp/cprogram.c", "pasp/ground.h",
                                "pasp/carray.c", "bitvector/bitvector.c"],
                     sources = ["pasp/ground.c", "pasp/cutils.c", "pasp/carray.c",
                                "pasp/cprogram.c", "bitvector/bitvector.c"],
                     include_dirs = [np.get_include()],
                     define_macros = STD_MACROS)
learn    = Extension("learn",
                     libraries = ["clingo", "pthread"],
                     depends = ["pasp/cprogram.c", "pasp/cinf.c", "pasp/cutils.c", "pasp/carray.c",
                                "pasp.ground.c", "pasp/cexact.c", "pasp/clearn.c", "pasp/cdata.c"],
                     sources = ["pasp/learn.c", "thpool/thpool.c", "pasp/cinf.c", "pasp/cprogram.c",
                                "bitvector/bitvector.c", "pasp/cutils.c", "pasp/clearn.c",
                                "pasp/carray.c", "pasp/cdata.c", "pasp/cexact.c", "pasp/coptimize.c"],
                     include_dirs = [np.get_include()],
                     extra_compile_args = ["-Wno-unused-function"],
                     define_macros = STD_MACROS)

sample   = Extension("sample",
                     libraries = ["clingo", "pthread"],
                     depends = ["pasp/cprogram.c", "pasp/cinf.c", "pasp/cutils.c", "pasp/carray.c",
                                "pasp.ground.c", "pasp/csample.c"],
                     sources = ["pasp/sample.c", "thpool/thpool.c", "pasp/cinf.c", "pasp/cprogram.c",
                                "bitvector/bitvector.c", "pasp/cutils.c", "pasp/csample.c",
                                "pasp/carray.c"],
                     include_dirs = [np.get_include()],
                     extra_compile_args = ["-Wno-unused-function"],
                     define_macros = STD_MACROS)

setup(
  packages = find_packages(where = ".", include = ["pasp*"]),
  include_package_data = True,
  ext_modules = [exact, ground, learn, sample],
  cmdclass = {"test": TestCommand},
)
